unit main;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ExtCtrls,
  ComCtrls, ActnList, Menus, StdActns, Buttons,
  fpstypes, fpspreadsheet, fpspreadsheetctrls, fpspreadsheetgrid, fpsActions,
  fpsallformats;

type

  { TForm1 }

  TForm1 = class(TForm)
    ExpandRange: TAction;
    ActionList1: TActionList;
    FileExit1: TFileExit;
    FileOpen1: TFileOpen;
    FileSaveAs1: TFileSaveAs;
    ImageList1: TImageList;
    MainMenu1: TMainMenu;
    MnuFile: TMenuItem;
    MnuFileOpen: TMenuItem;
    MnuFileSaveAs: TMenuItem;
    MnuFileSeparator: TMenuItem;
    MnuFileExit: TMenuItem;
    MnuFormat: TMenuItem;
    MnuFormatBold: TMenuItem;
    MnuFormatItalic: TMenuItem;
    MnuFormatUnderline: TMenuItem;
    sWorkbookSource1: TsWorkbookSource;
    sWorkbookTabControl1: TsWorkbookTabControl;
    sWorksheetGrid1: TsWorksheetGrid;
    sCellEdit1: TsCellEdit;
    sCellIndicator1: TsCellIndicator;
    sFontStyleAction1: TsFontStyleAction;
    sFontStyleAction2: TsFontStyleAction;
    sFontStyleAction3: TsFontStyleAction;
    Panel2: TPanel;
    FontNameCombo: TsCellCombobox;
    FontSizeCombo: TsCellCombobox;
    Splitter2: TSplitter;
    Splitter3: TSplitter;
    ToolBar3: TToolBar;
    ToolBar1: TToolBar;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    ToolButton4: TToolButton;
    ToolButton5: TToolButton;
    ToolButton6: TToolButton;
    ToolButton7: TToolButton;
    ToolButton8: TToolButton;
    procedure ActionList1Execute(AAction: TBasicAction; var Handled: Boolean);
    procedure FileOpen1Accept(Sender: TObject);
    procedure FileSaveAs1Accept(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure SelectCurrentRangeExecute(Sender: TObject);
    function getCurrentRange(ws:TsWorksheet;ARow,ACol:Cardinal):TsCellRange;
  private

  protected

  public

  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

uses
  LCLIntf, fpsUtils, fpsCSV;

{ TForm1 }

{Two procedures to describe the problem}

{Attempt to expand the active cell to the full current range -
a rectangular area bounded by rows and columns without visible
text (not diagonal, just directly up-down-left-right - not as
in Excel or Calc) }
procedure TForm1.SelectCurrentRangeExecute(Sender: TObject);
var ws:TsWorksheet;
  allSelRange:TsCellRangeArray;
  SeleRange:TsCellRange;
  activeCellRow,activeCellCol:Cardinal;
begin
  ws:=sWorkbookSource1.Workbook.ActiveWorksheet;
  activeCellRow:=ws.ActiveCellRow;
  activeCellCol:=ws.ActiveCellCol;
{ !!! Here first problem: getCurrentRange work properly for many cells,
but for other (E5, H7, J20...) returns an incorrect result -
or a range of one start cell, or [1,12] !!! }

  SeleRange:=getCurrentRange(ws,activeCellRow,activeCellCol);
  ShowMessageFmt('Active cell [%D,%D]'+ #13+'Current range [%D,%D][%D,%D]',
    [activeCellRow, activeCellCol,
    SeleRange.Row1,SeleRange.Col1,SeleRange.Row2,SeleRange.Col2]);
  SetLength(allSelRange,1);
  allSelRange[0]:=SeleRange;
{ !!! Second problem here - What else we need to do to draw a range on the screen? !!! }
  ws.SetSelection(allSelRange);
  allSelRange := ws.GetSelection;
  ShowMessageFmt('Set the selection for the range [%D,%D][%D,%D]' + #13
  +'Current selection has %D areas' + #13
  +'First area is the range [%D,%D][%D,%D]',
  [ SeleRange.Row1,SeleRange.Col1,SeleRange.Row2,SeleRange.Col2,
  ws.GetSelectionCount, allSelRange[0].Row1,allSelRange[0].Col1,
  allSelRange[0].Row2,allSelRange[0].Col2]);

end;

function TForm1.getCurrentRange(ws:TsWorksheet; ARow, ACol:Cardinal): TsCellRange;
var
  FirstCol,LastCol,FirstRow,LastRow:Cardinal; { Borders of UsedRange }
  Range:TsCellRange; { Start from Cell(ARow, ACol) and expand it if possible }
  alreadyVisit: array of array of Boolean; { Map of UsedRange - mark visited cells }

  procedure extendSelection(const ARow,ACol:Cardinal);
  begin
      if (ARow<FirstRow)
         or (ARow>LastRow)
         or (ACol< FirstCol)
         or (ACol > LastCol) then Exit; { Outbound }
      if alreadyVisit[ARow-FirstRow][ACol-FirstCol] then Exit; { Not repeat twice }
      alreadyVisit[ARow-FirstRow][ACol-FirstCol]:=True; { Mark as "Was here" }
  { !!! Error position: sometimes the method returns an empty string because it
   does not find the specified existing cell !!! }
      if ws.ReadAsText(ARow,ACol)='' then Exit;
  { Resize result range if need }
      if Range.Col1>ACol then Range.Col1:=ACol;
      if Range.Col2<ACol then Range.Col2:=ACol;
      if Range.Row1>ARow then Range.Row1:=ARow;
      if Range.Row2<ARow then Range.Row2:=ARow;
      if (Range.Col1<=FirstCol)
         and (Range.Col2 >= LastCol)
         and (Range.Row2=LastRow)
         and (Range.Row1=FirstRow) then Exit; // Range expanded to UsedRange, no sense to continue
  { Check adjacent cells }
      if ARow >= FirstRow then extendSelection(ARow-1,ACol); // To Up
      if ACol <= LastCol then extendSelection(ARow,ACol+1); // To Right
      if ARow <= LastRow then extendSelection(ARow+1,ACol); // To Down
      if ACol >= FirstCol then extendSelection(ARow,ACol-1); // To Left
  end;

begin
  FirstRow:=ws.GetFirstRowIndex();
  FirstCol:=ws.GetFirstColIndex();
  LastRow:=ws.GetLastRowIndex();
  LastCol:=ws.GetLastColIndex();
  SetLength(alreadyVisit, LastRow-FirstRow+1,LastCol-FirstCol+1);
  Range.Col1:= ACol;
  Range.Col2:= ACol;
  Range.Row1:= ARow;
  Range.Row2:= ARow;
{ Start from the specified cell }
  extendSelection(ACol,ARow);
  Result:=Range;

  SetLength(alreadyVisit,0);
end;

{ Rest code same as demo_ctrls }
{ Loads the spreadsheet file selected by the FileOpen standard action }
procedure TForm1.FileOpen1Accept(Sender: TObject);
begin
  sWorkbookSource1.AutodetectFormat := false;
  case FileOpen1.Dialog.FilterIndex of
    1: sWorkbookSource1.AutoDetectFormat := true;         // All spreadsheet files
    2: sWorkbookSource1.AutoDetectFormat := true;         // All Excel files
    3: sWorkbookSource1.FileFormat := sfOpenDocument;     // Open/LibreOffice
  end;
  sWorkbookSource1.FileName :=FileOpen1.Dialog.FileName;  // This loads the file
end;

procedure TForm1.ActionList1Execute(AAction: TBasicAction; var Handled: Boolean
  );
begin
  SelectCurrentRangeExecute(Form1);
end;

{ Saves the spreadsheet to the file selected by the FileSaveAs1 action }
procedure TForm1.FileSaveAs1Accept(Sender: TObject);
var
  fmt: TsSpreadsheetFormat;
begin
  Screen.Cursor := crHourglass;
  try
    case FileSaveAs1.Dialog.FilterIndex of
      1: fmt := sfOOXML;
      2: fmt := sfOpenDocument;
    end;
    sWorkbookSource1.SaveToSpreadsheetFile(FileSaveAs1.Dialog.FileName, fmt);
  finally
    Screen.Cursor := crDefault;
  end;
end;

procedure TForm1.FormShow(Sender: TObject);
begin
sWorkbookSource1.FileName:=ExtractFilePath(ApplicationName) + 'TestRangeBook.ods';
end;

end.

